// ----------------------------------------------------------------------------
//
// rt.cc
// RT++ --- higher order threads for C++
//
// $Id: rt.cc,v 1.41 1996/04/03 08:33:25 ws Exp $
// (C) 1996 Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
//
// ----------------------------------------------------------------------------

#include "rt++.h" 

static jmp_buf caller; // state for non-local exit
static int code;       // return code

static void *sig_int;  // save buffers for original signal handlers
static void *sig_term;
static void *sig_segv;
static void *sig_bus;
static void *sig_hup;
static void *sig_chld;
static void *sig_usr1;

// ----------------------------------------------------------------------------
//
// trigger shutdown of program
//
// ----------------------------------------------------------------------------
static void 
shutdown(int sig) 
{
  if (sig == SIGCHLD)             // child has terminated
    {
      cout << "\nRT++ aborted: child has terminated.\n";
    }
  else if (sig == SIGINT)         // ctrl-c 
    {
      cout << "\nRT++ aborted on user request.\n"; 
    }
  else if (sig != SIGHUP)         // non-proper termination
    {
      cout << "\nRT++ process " << processor.number() << ": signal " 
	<< sig << " received (core dumped).\n"; 
    }
  code = sys->terminate();        // terminate program
  rt_stat->report();              // report statistics
  if (sig == SIGCHLD)
    {
      exit(code);                 // error exit if child has died
    }
  if (sig == SIGINT)
    {
      exit(code);                 // error exit on ctrl-c
    }
  if (sig != SIGHUP) 
    {
      abort();                    // produce core and exit
    }
  signal(SIGINT, SA_HANDLER(sig_int));   // ctrl c
  signal(SIGTERM, SA_HANDLER(sig_term)); // kill
  signal(SIGSEGV, SA_HANDLER(sig_segv)); // segmentation violation
  signal(SIGBUS, SA_HANDLER(sig_bus));   // bus error
  signal(SIGCHLD, SA_HANDLER(sig_chld)); // death of child
  signal(SIGHUP, SA_HANDLER(sig_hup));   // termination triggered
  signal(SIGUSR1, SA_HANDLER(sig_usr1)); // garbage collection
  longjmp(caller, 1);                    // non-local jump to saved state
}

// ----------------------------------------------------------------------------
//
// dump core and abort
//
// ----------------------------------------------------------------------------
static void 
coredump(int sig) 
{
  cout << "\nRT++ process " << processor.number() << ": signal "
    << sig << " received (core dumped).\n"; 
  abort();
}

// ----------------------------------------------------------------------------
//
// initial thread to be executed
//
// ----------------------------------------------------------------------------
class BaseThread: public CoreThread
{
  int (*f)(int, char *[]);   // user main program
  int argc;                  // argument count
  char **argv;               // argument vector
  void user(void)            // code to be executed
    {
      rt_bind(0);
      rt_exit(f(argc, argv));
    }
  public:
  void scan(void)            // scan function for garbage collection
    {
      CoreThread::scan();
    }
  void init(int (*_f)(int, char *[]), int _argc, char *_argv[])
    {
      CoreThread::init();
      f = _f;
      argc = _argc;
      argv = _argv;
      userfun = (void(*)(CoreThread*))user;
      scanfun = (void(*)(CoreThread*))BaseThread::scan;
    }
  BaseThread(int (*_f)(int, char*[]), int _argc, char *_argv[])
    {
      init(_f, _argc, _argv);
    }
};

class InitThread
{
  BaseThread *pointer;
  public:
  void init(int (*_f)(int, char*[]), int _argc, char *_argv[])
    {
      // must be realized for binding to take effect
      pointer = (BaseThread*)processor.alloc(sizeof(BaseThread));
      pointer->init(_f, _argc, _argv);
      pointer->realize();
      pointer->root();
      sys->is_ready(pointer);
      processor.unlock();
    }
  InitThread(int (*_f)(int, char*[]), int _argc, char *_argv[])
    {
      init(_f, _argc, _argv);
    }
};

Ref(BaseThread);

// ----------------------------------------------------------------------------
//
// execute function <f> with threads and deliver return code 
// use <n> processors for execution
// use <m> bytes for shared memory
// use <s> bytes for thread stack
// use <h> bytes for heap
// use <v> as verbosity level
//
// ----------------------------------------------------------------------------
int
rt_main(int (*f)(int, char *[]), int argc, char *argv[],
	const RT_Argument &rt_arg, RT_Result *rt_res)
{
  // continuation state for program exit
  if (setjmp(caller)) 
    {
      shm_shut();   // shutdown shared memory
      return(code); // return to caller        
    }

  // argument and result structure
  rt_argument = rt_arg;
  rt_result = rt_res;

  // initialize shared memory
  shm_init(rt_arg.memory);

  // initialize process system
  sys_init(rt_arg.proc, rt_arg.stack);
        
  // initialize garbage-collected heap
  heap_init(rt_arg.heap, rt_arg.blocks);

  // initialize statistics
  stat_init(rt_arg.verbose);

  const int n = rt_arg.proc;         // number of processors
  pid_t pids[n];
  pids[0] = getpid();

  rt_shm_unmap();                    // unmap shared memory
  rt_fork_start();                   // start forking processes
  for (int i = 1; i < n; i++)        // create new Unix processes
    {				    
      pid_t pid = rt_fork();         // Unix process identifier
      if (pid == 0)                  // child process
	{
	  signal(SIGSEGV, SA_HANDLER(coredump)); // segmentation violation
	  signal(SIGBUS, SA_HANDLER(coredump));  // coredump on bus error
	  rt_shm_map();                          // map shared memory
	  processor.init(i);                     // initialize process
	  sys->wait();                           // global synchronization
	  processor.scheduler();                 // never returns
	}
      if (pid == -1) rt_abort("Unix process creation failed");
      pids[i] = pid;                 // register process identifier
    }
  rt_fork_end();                     // end forking processes
  rt_shm_map();                      // map shared memory

  sig_int = signal(SIGINT, SA_HANDLER(shutdown));   // ctrl c
  sig_term = signal(SIGTERM, SA_HANDLER(shutdown)); // kill
  sig_segv = signal(SIGSEGV, SA_HANDLER(shutdown)); // segmentation violation
  sig_bus = signal(SIGBUS, SA_HANDLER(shutdown));   // bus error
  sig_chld = signal(SIGCHLD, SA_HANDLER(shutdown)); // death of child
  sig_hup = signal(SIGHUP, SA_HANDLER(shutdown));   // termination triggered
  sig_usr1 = signal(SIGUSR1, SIG_IGN);              // gc signal

  for (int j = 0; j < n; j++)
    sys->proc(j, pids[j]);           // register process identifier

  processor.init(0);		     // init process
  sys->wait();                       // global synchronization
  rt_shm_close();                    // close shared memory
  rt_stat->start();                  // initialize timer
  InitThread(f, argc, argv);         // create initial thread
  processor.scheduler();             // execute thread scheduler 
  return(0);                         // never reached
}

// ----------------------------------------------------------------------------
//
// terminate program with return code 
//
// ----------------------------------------------------------------------------
void 
rt_exit(int code)
{
  sys->lock();         // prevent termination by other threads
  sys->shutdown(code); // shutdown system and deliver exit code
}

// ----------------------------------------------------------------------------
//
// abort program with error message
//
// ----------------------------------------------------------------------------
void 
rt_abort(char *msg)
{
  sys->lock();                    // prevent termination by other threads
  cout << "\nRT++ aborted: " << msg << ".\n";
  sys->shutdown(-1);              // shutdown system and deliver exit code -1
}

// ----------------------------------------------------------------------------
//
// abort program with error message and immediate exit
//
// ----------------------------------------------------------------------------
void
rt_panic(char *msg)
{
  cout << "\nRT++ paniced: " << msg << ".\n";
  abort();                        // produce core
}

// ----------------------------------------------------------------------------
// end of rt.cc
// ----------------------------------------------------------------------------
// Local Variables: ***
// mode: C++ ***
// End: ***
