// ----------------------------------------------------------------------------
//
// rt_core.cc
// RT++ core threads (untyped)
//
// $Id: rt_core.cc,v 1.32 1996/04/03 08:33:26 ws Exp $
// (C) 1996 Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
//
// ----------------------------------------------------------------------------

#include "rt++.h"

// ----------------------------------------------------------------------------
//
// actual function executed by thread
//
// ----------------------------------------------------------------------------
static void 
shell(void *t, void *, void(*)(void*))
{
  CoreThread *thread = (CoreThread*)t; // this is the current thread
  thread->userfun(thread);             // execute user function
  thread->deliver();                   // deliver result
  processor.poll() ;                   // execute other lazy threads
}

// ----------------------------------------------------------------------------
//
// realize thread
//
// ----------------------------------------------------------------------------
void 
CoreThread::realize(void)
{
  executing = 1;                                    // flag execution
  int size = sys->stacksize();                      // the stack size
  void *stack = SystemThread::realize(size, shell); // allocate stack
  shadow.init(stack);                               // initialize shadow stack
}

// ----------------------------------------------------------------------------
//
// return result of thread
//
// ----------------------------------------------------------------------------
void
CoreThread::wait(void) 
{
  processor.lock();                    // prevent preemption
  lock();                              // lock thread
  if (avail)                           // result available
    {				      
      unlock();                        // unlock thread
      processor.unlock();              // unlock processor
      return;                          // we are done
    }
  CoreThread *thread = processor.get_thread();
  if (executing)                       // thread already executing
    {
      thread->next = waiters;          // link thread into queue
      waiters = thread;
      processor.block(&mutex);         // block current thread and unlock
      return;                          // we are done
    }
  executing = 1;                       // thread is executing
  unlock();                            // unlock thread
  sys->get_lazy(this);                 // unlink thread from lazy list
  this->next = thread->lazy;           // push thread on thread lazy list
  thread->lazy = this;                 // critical region for gc!!!
  this->lazy = thread;
  processor.unlock();                  // unlock processor
  thread->check();                     // check for stack overflow
  userfun(this);                       // execute user code
  deliver();                           // deliver result
  thread->lazy = thread->lazy->next;   // pop thread from thread lazy list
  this->lazy = 0;
}

// ----------------------------------------------------------------------------
//
// stop thread i.e. prevent execution or signal end of execution
// caller is in charge of locking processor!
// processor remains unlocked!
//
// ----------------------------------------------------------------------------
void
CoreThread::_stop(void) 
{
  lock();                              // lock thread
  if (avail)                           // result available
    {				      
      unlock();                        // unlock thread
      return;                          // we are done
    }
  if (executing)                       // thread already executing
    {
      stopped = 1;                     // signal stopping of thread
    }
  else
    {
      executing = 1;                   // pretend thread is executing
    }
  unlock();                            // unlock thread
}

// ----------------------------------------------------------------------------
//
// stop thread i.e. prevent execution or signal end of execution
//
// ----------------------------------------------------------------------------
void
CoreThread::stop(void) 
{
  processor.lock();
  _stop();
  processor.unlock();
}

// ----------------------------------------------------------------------------
//
// deliver thread result to blocked threads
// zero thread arguments such that they are not referenced in gc any more
// processor remains locked!!!
//
// ----------------------------------------------------------------------------
void
CoreThread::deliver(void)
{
  if (zerofun) zerofun(this);
  processor.lock();
  lock();        
  avail = 1;     
  while (waiters)
    {
      CoreThread *thread = waiters;
      waiters = thread->next;
      sys->is_ready(thread);
    }
  unlock();
}

// ----------------------------------------------------------------------------
//
// mark thread for execution such that it is not executed by other thread
// return 1 if thread not yet marked by other thread
// caller is in charge of locking the processor!!!
//
// ----------------------------------------------------------------------------
int
CoreThread::execute(void)
{
  lock();            // lock thread
  if (executing)     // thread already executing
    {		     
      unlock();      // unlock thread
      return(0);     // failure case
    }		     
  else		     
    {		     
      executing = 1; // set thread for execution
      unlock();      // unlock thread
      return(1);     // success case
    }
}

// ----------------------------------------------------------------------------
//
// bind thread to specific processor <n>
//
// ----------------------------------------------------------------------------
void
CoreThread::bind(int n)
{
  bound = n;
}

// ----------------------------------------------------------------------------
//
// unbind thread
//
// ----------------------------------------------------------------------------
void
CoreThread::unbind(void)
{
  bound = -1;
}

// ----------------------------------------------------------------------------
//
// declare thread as root of garbage collection
//
// ----------------------------------------------------------------------------
void
CoreThread::root(void)
{
  processor.lock();
  sys->root_thread(this);
  processor.unlock();
}

// ----------------------------------------------------------------------------
//
// reset declaration
//
// ----------------------------------------------------------------------------
void
CoreThread::noroot(void)
{
  processor.lock();
  sys->no_root_thread(this);
  processor.unlock();
}

// ----------------------------------------------------------------------------
// end of rt_core.cc
// ----------------------------------------------------------------------------
// Local Variables: ***
// mode: C++ ***
// End: ***
