// ----------------------------------------------------------------------------
//
// rt_heap.cc
// RT++ garbage collected heap
//
// $Id: rt_heap.cc,v 1.27 1996/04/03 15:43:17 ws Exp ws $
// (C) 1996 Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
//
// ----------------------------------------------------------------------------

#include "rt++.h"

Heap *heap; // the heap object

// ----------------------------------------------------------------------------
//
// initialize heap
//
// ----------------------------------------------------------------------------
void 
heap_init(int size, int blocks)
{
  heap = (Heap*)shm_alloc(sizeof(Heap));
  heap->Heap(size, blocks);
}

// ----------------------------------------------------------------------------
//
// allocate new cluster and link last block with <next>
//
// ----------------------------------------------------------------------------
FreeBlock *
Heap::cluster(FreeBlock *next)
{
  int clustersize = sizeof(FreeBlock)+size;
  void *memory = shm_alloc(clustersize+rt_alignment()-1);
  void *aligned = rt_align(memory+sizeof(FreeBlock)+sizeof(Block));
  FreeBlock *cluster = (FreeBlock*)(aligned-sizeof(Block)-sizeof(FreeBlock));
  void *clusterend = ((void*)cluster)+clustersize;
  cluster->size = clustersize;
  for (void *ptr = cluster+1; ptr < clusterend; ptr += blocksize)
    {
      FreeBlock *block = (FreeBlock*)ptr;
      block->size = blocksize;
      block->next = (FreeBlock*)(ptr+blocksize);
    }
  ((FreeBlock*)(clusterend-blocksize))->next = next;
  return(cluster);
}

// ----------------------------------------------------------------------------
//
// create heap of one cluster of at least <s> bytes
// the cluster (and each additional one) is split into <b> blocks per processor
//
// ----------------------------------------------------------------------------
Heap::Heap(int s, int b)
{
  int procnum = sys->procnum();
  blocknum = procnum*b;
  s += (blocknum - s%blocknum) % blocknum;
  blocksize = (int)rt_align((void*)Block::align(s/blocknum));
  size = blocknum*blocksize;
  start = cluster(0);
  free = start+1;
  notready = sys->procnum();
  triggered = 0;
  stacks = 0;
}

// ----------------------------------------------------------------------------
//
// return heap size
//
// ----------------------------------------------------------------------------
int
Heap::heapsize(void)
{
  return(size);
}

// ----------------------------------------------------------------------------
//
// return number of heap blocks
//
// ----------------------------------------------------------------------------
int
Heap::blocknumber(void)
{
  return(blocknum);
}

// ----------------------------------------------------------------------------
//
// trigger garbage collection
//
// ----------------------------------------------------------------------------
void
Heap::collect(void)
{
  processor.lock();                 
  mutex_alloc.lock();
  if (triggered)                        // gc already triggered
    {
      mutex_alloc.unlock();
      processor.unlock();               // give gc signal a chance
      while (triggered) { }             // stay in user mode
      return;
    }
  triggered = 1;                        // trigger gc
  trigger = processor.number();         // remember that we triggered
  triggersize = 0;                      // remember request size
  mutex_alloc.unlock();                 // release blocked processors
  rt_stat->gc_trigger();                // report request for gc
  sys->gc();                            // put other procs in gc state
  processor.unlock();                   // give gc signal a chance
  raise(SIGUSR1);                       // bring us into gc state
  while (triggered) { }                 // stay in user mode
}

// ----------------------------------------------------------------------------
//
// return memory block cycle and store predecessor of first block in <prev>
// if not available, trigger garbage collection and return 0
// if largest block after gc is less than <s> bytes, increase heap or abort
// the caller is in charge of locking the processor!!!
// if 0 is returned, the processor was unlocked!!!
//
// ----------------------------------------------------------------------------
FreeBlock*
Heap::alloc(int s, FreeBlock **prev)
{
  if (s > blocksize)
    {
      rt_abort("too much memory requested; please increase blocksize");
    }
  mutex_alloc.lock();                       
  if (triggered)                        // gc already triggered
    {
      mutex_alloc.unlock();
      processor.unlock();               // give gc signal a chance
      while (triggered) { }             // stay in user mode
      return(0);                        // signal failure
    }
  FreeBlock *last = 0;                  // predecessor block
  FreeBlock *block = free;              // first block
  int n = blocksize;                    // minimum return size
  while (n > 0)                         // construct list of appr size
    {
      if (!free) break;                 // free list consumed
      n -= free->size;                  // one more block
      last = free;                      // go to next block
      free = free->next;
    }
  if (n <= 0)                           // list okay
    {
      last->next = block;               // make cycle out of block
      *prev = last;                     // predecessor of first block
      mutex_alloc.unlock();             // release blocked processors
      return(block);                    // success
    }
  triggered = 1;                        // trigger gc
  trigger = processor.number();         // remember that we triggered
  triggersize = s;                      // remember request size
  mutex_alloc.unlock();                 // release blocked processors
  rt_stat->gc_trigger();                // report request for gc
  sys->gc();                            // put other procs in gc state
  processor.unlock();                   // give gc signal a chance
  raise(SIGUSR1);                       // bring us into gc state
  while (triggered) { }                 // stay in user mode
  return(0);                            // signal failure
}

// ----------------------------------------------------------------------------
//
// mark phase of garbage collection
// each stack is marked by a processor,
// stacks of referenced threads are pushed on the stacks list
//
// ----------------------------------------------------------------------------
void 
Heap::mark(void)
{
  while (1)                                   // mark root threads
    {
      CoreThread *thread;
      mutex_gc.lock();
      thread = threads;
      if (thread)
	{
	  threads = thread->root_next;
	}
      mutex_gc.unlock();
      if (!thread) break;
      if (!Block::block(thread)->mark())
	{
	  thread->scanfun(thread);
	}
    }
  while (1)                                  // mark stacks of marked threads
    {
      mutex_gc.lock();
      Stack *stack = (Stack*)stacks;
      if (stack) stacks = stack->next;
      mutex_gc.unlock();
      if (stack) 
	{
	  stack->mark();
	  stack->gc = 0;
	  continue;
	}
      mutex_gc.lock();
      _markers--;
      mutex_gc.unlock();
      while (!stacks && _markers > 0) { }
      if (!stacks) break;
      mutex_gc.lock();
      _markers++;
      mutex_gc.unlock();
    }
}

// ----------------------------------------------------------------------------
//
// sweep phase of garbage collection
// each sector (original free block) is swept by a processor
//
// ----------------------------------------------------------------------------
void 
Heap::sweep(void)
{
  FreeBlock *first = 0; // first reclaimed block in sector
  FreeBlock *last = 0;  // last reclaimed block in sector
  int mysize = 0;       // total size of reclaimed blocks in sector
  int mymax = 0;        // largest block in sector
  int myblocks = 0;     // total number of reclaimed blocks in sector
  while (1)
    {
      mutex_gc.lock();         // get sector
      Block *next = blocks;
      if (next == clusterend)
	{
	  if (clusters)
	    {
	      next = clusters+1;
	      clusterend = (Block*)(((void*)next) + size);
	      clusters = clusters->next;
	    }
	  else
	    {
	      mutex_gc.unlock();
	      break;
	    }
	}
      Block *blockend = (Block*)(((void*)next) + blocksize);
      blocks = blockend;
      mutex_gc.unlock();
      while (1)                // reclaim blocks in sector
	{
	  while (next < blockend && next->unmark())
	    next = (Block*)(((void*)next) + next->size);
	  if (next == blockend) break;
	  FreeBlock *block = (FreeBlock*)next;
	  do
	    {
	      next = (Block*)(((void*)next) + next->size);
	    }
	  while (next < blockend && !next->is_marked());
	  int size = (void*)next-(void*)block;
	  block->size = size;
	  if (first)
	    last->next = block;
	  else
	    first = block;
	  last = block;
	  mysize += size;
	  if (size > mymax) mymax = size;
	  myblocks++;
	}
    }
  if (first)                  // put reclaimed blocks into free list
    {
      mutex_gc.lock();
      last->next = free;
      free = first;
      freesize += mysize;
      freeblocks += myblocks;
      if (mymax > maxsize) maxsize = mymax;
      mutex_gc.unlock();
    }
}

// ----------------------------------------------------------------------------
//
// perform garbage collection
//
// ----------------------------------------------------------------------------
void 
Heap::gc(void)
{
  processor.gc();                           // prepare processor for gc
  int number = processor.number();          // our processor

  if (number == trigger)                    // prepare gc
    {
      int procnum = sys->procnum();
      markers = procnum;
      _markers = procnum;
      sweepers = procnum;
      while (notready > 1) { }              // synchronize with other procs
      threads = sys->root_threads();
      clusters = start;
      blocks = 0;
      clusterend = 0;
      free = 0;
      freesize = 0;
      freeblocks = 0;
      maxsize = 0;
      rt_stat->gc_start();                  // report start of gc
    }

  mutex_gc.lock();                          // synchronize before marking
  notready--;
  mutex_gc.unlock();
  while (notready > 0) { }

  mark();                                   // mark phase

  if (number == trigger)                    // sweep thread lists
    {
      while (markers > 1) { }               // synchronize with other procs
      sys->gc_threads();                    // collect unreferenced threads
    }

  mutex_gc.lock();                          // synchronize before sweeping
  markers--;
  mutex_gc.unlock();
  while (markers > 0) { }

  sweep();                                  // sweep phase

  if (number == trigger)                    // cleanup gc
    {
      while (sweepers > 1) { }              // synchronize with other procs
      rt_stat->gc_end(freeblocks, freesize);// report end of gc
      const int minimum = sys->procnum()*blocksize;
      if (freesize < minimum) 
	{
	  rt_stat->gc_free(minimum);
	  gc_inc();
	}
      else if (maxsize < triggersize) 
	{
	  rt_stat->gc_size(triggersize);
	  gc_inc();
	}
      notready = sys->procnum();            // prepare for next gc
      triggered = 0;                        // show that gc is completed
    }

  mutex_gc.lock();                          // for synchronization of trigger
  sweepers--;
  mutex_gc.unlock();
  while (sweepers > 0) { }
}

// ----------------------------------------------------------------------------
//
// extend heap by new cluster
//
// ----------------------------------------------------------------------------
void
Heap::gc_inc(void)
{
  rt_stat->gc_inc(size);        // report to system
  FreeBlock *c = cluster(free); // allocate new cluster and link with free list
  free = c+1;                   // new free list
  c->next = start;              // link cluster into cluster list
  start = c;
}

// ----------------------------------------------------------------------------
// end of rt_heap.cc
// ----------------------------------------------------------------------------
// Local Variables: ***
// mode: C++ ***
// End: ***
