// ----------------------------------------------------------------------------
//
// rt_proc.cc
// RT++ processor-specific data
//
// $Id: rt_proc.cc,v 1.46 1996/04/03 15:43:20 ws Exp ws $
// (C) 1996 Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
//
// ----------------------------------------------------------------------------

#include "rt++.h"

// static class variables, only defined once in current process

int Processor::kernel;           // set if thread executes in kernel mode
CoreThread *Processor::thread;   // currently executed thread
Stack Processor::shadow;         // shadow stack
CoreThread *Processor::sched;    // scheduler thread

int Processor::num;              // processor number
FreeBlock *Processor::free;      // free block
int Processor::size;             // size of free block
FreeBlock *Processor::prev;      // predecessor block

sigset_t Processor::empty_mask;  // empty signal mask
sigset_t Processor::alarm_mask;  // mask with SIGALRM
sigset_t Processor::gc_mask;     // mask with SIGALRM and SIGUSR1

int Processor::gc_signal;        // garbage collection signal

Processor processor;             // statically allocated in local data segment

static void preempt(int);        // called on preemption of thread
static void gc(int);             // called on garbage collection

// ----------------------------------------------------------------------------
//
// switch off preemption
//
// ----------------------------------------------------------------------------
void
Processor::nopreempt(void)
{
  sigprocmask(SIG_BLOCK, &alarm_mask, 0);
}

// ----------------------------------------------------------------------------
//
// switch on preemption
//
// ----------------------------------------------------------------------------
void
Processor::preempt(void)
{
  sigprocmask(SIG_UNBLOCK, &alarm_mask, 0);
}

// ----------------------------------------------------------------------------
//
// switch off gc
//
// ----------------------------------------------------------------------------
void
Processor::gc_off(void)
{
  sigprocmask(SIG_BLOCK, &gc_mask, 0);
}

// ----------------------------------------------------------------------------
//
// switch on preemption
//
// ----------------------------------------------------------------------------
void
Processor::gc_on(void)
{
  sigprocmask(SIG_UNBLOCK, &gc_mask, 0);
}

// ----------------------------------------------------------------------------
//
// set handler for alarm timer (preempt thread)
// SIGALRM must be blocked during execution of the handler!!!
//
// ----------------------------------------------------------------------------
static inline void 
setalarm(void) 
{
  struct sigaction action;
  action.sa_handler = SA_HANDLER(preempt);
  sigemptyset(&action.sa_mask);
  action.sa_flags = 0;
  int ret = sigaction(SIGALRM, &action, 0);
  if (ret < 0) rt_abort("installation of alarm handler failed");
}

// ----------------------------------------------------------------------------
//
// ignore SIGALRM and schedule it again
//
// ----------------------------------------------------------------------------
static void
ignore(void)
{
  alarm(1);
}

// ----------------------------------------------------------------------------
//
// set handler for alarm timer (ignore signal)
// SIGALRM must be blocked during execution of the handler!!!
//
// ----------------------------------------------------------------------------
static inline void 
setnoalarm(void) 
{
  struct sigaction action;
  action.sa_handler = SA_HANDLER(ignore);
  sigemptyset(&action.sa_mask);
  action.sa_flags = 0;
  int ret = sigaction(SIGALRM, &action, 0);
  if (ret < 0) rt_abort("installation of alarm handler failed");
}

// ----------------------------------------------------------------------------
//
// preempt current thread and reschedule SIGALRM
// SIGALRM is blocked during execution of the handler!!!
//
// ----------------------------------------------------------------------------
static void 
preempt(int) 
{
  alarm(1);                                     // reset alarm timer
  if (processor.locked())                       // processor in kernel mode
    {
      return;
    }
  processor.lock();                             // prevent further preemption
  CoreThread *next = sys->get_ready();          // get ready thread
  if (next)                                     // thread was found
    {
      processor.save();                         // going to switch thread
      processor.preempt();                      // return of sig handler
      next->exec_ready();	                // non-local jump
      processor.nopreempt();                    // in sig handler again
    }
  else
    {
      processor.unlock();                       // exit kernel mode
    }
}

// ----------------------------------------------------------------------------
//
// set gc handler
// SIGUSR1 (gc signal) and SIGALRM must be blocked during handler execution!
//
// ----------------------------------------------------------------------------
static inline void 
setgc(void) 
{
  struct sigaction action;
  action.sa_handler = SA_HANDLER(gc);
  sigemptyset(&action.sa_mask);
  sigaddset(&action.sa_mask, SIGALRM);
  action.sa_flags = 0;
  int ret = sigaction(SIGUSR1, &action, 0);
  if (ret < 0) rt_abort("installation of garbage collector failed");
}

// ----------------------------------------------------------------------------
//
// switch to scheduler to perform gc
//
// ----------------------------------------------------------------------------
void
Processor::collect(void)
{
  if (thread == sched)      // scheduler may directly do gc
    {
      heap->gc();           // go into gc state
    }
  else                      // let thread go into gc state
    {
      CoreThread *t = processor.get_thread();   // current thread
      int b = t->binding();                     // current binding
      if (b == -1) t->bind(processor.number()); // must return to this process
      gc_signal = 1;                            // set gc signal
      save();                                   // going to switch thread
      lock();                                   // enter kernel mode
      gc_on();                                  // return from sig handler
      sched->exec_ready();                      // non-local jump to scheduler
      gc_off();                                 // in sig handler again
      gc_signal = 0;                            // reset gc signal
      if (b == -1) t->unbind();                 // restore binding
    }
}

// ----------------------------------------------------------------------------
//
// execute on receipt of garbage collection signal
// SIGUSR1 (gc signal) and SIGALRM are blocked during handler execution!
//
// ----------------------------------------------------------------------------
static void 
gc(int) 
{
  if (processor.locked())   // we cannot do garbage collection yet
    {
      sys->set(1);          // however report that we received gc signal
    }
  else
    {
      sys->set(2);          // acknowledge gc signal
      processor.collect();  // perform gc
    }
}

// ----------------------------------------------------------------------------
//
// initialize current processor with number <n>
//
// ----------------------------------------------------------------------------
void 
Processor::init(int n) 
{
  num = n;                                 // processor number
  kernel = 1;                              // in kernel mode
  gc_signal = 0;                           // no gc signal set
  sched = (CoreThread*)shm_alloc(sizeof(CoreThread));
  sched->init();                           // scheduler thread
  thread = sched;                          // current thread
  free = 0;                                // current free block
  prev = 0;                                // predecessor block
  size = 0;                                // size of current free block
  sigemptyset(&empty_mask);                // empty signal mask
  sigemptyset(&alarm_mask);                // mask with SIGALRM only
  sigaddset(&alarm_mask, SIGALRM);
  sigemptyset(&gc_mask);                   // mask with SIGALRM and SIGUSR1
  sigaddset(&gc_mask, SIGALRM);
  sigaddset(&gc_mask, SIGUSR1);
  setalarm();                              // set alarm timer
  setgc();                                 // set gc handler
  alarm(1);                                // schedule signal
}

// ----------------------------------------------------------------------------
//
// poll for ready thread and bring it to execution
// if no ready thread is found, realize lazy thread
// if no lazy thread is found, busy wait for some time
// after that, suspend process until timer signal
// caller is in charge of locking the processor!!!
//
// ----------------------------------------------------------------------------
void 
Processor::scheduler(void) 
{
  while (1)
    {
      CoreThread *ready = sys->get_ready();   // get ready thread
      if (!ready)                             // ready thread not found
	{                              
	  CoreThread *lazy = sys->get_lazy(); // get lazy thread
	  if (lazy) lazy->realize();          // realize lazy thread
	  ready = lazy;                       // new ready thread
	}
      if (ready)                              // ready thread found
	{
	  ready->exec_basic();	              // switch to thread
	  if (gc_signal)
	    {
	      gc_signal = 0;
	      heap->gc();                     // perform gc
	    }
	}
      else                                    // no thread available
	{       
	  nopreempt();                        // switch off SIGALRM
	  unlock();                           // allow gc signal to be answered
	  const int n = sys->poll();          // a fraction of an alarm slice
	  for (int i = 0; i < 10*n; i++)         // poll for threads
	    {                                 // may perform gc during loop
	      if (sys->thread_avail()) break;
	    }
	  while (!sys->thread_avail())
	    {
	      setnoalarm();                   // return on SIGALRM
	      sigsuspend(&empty_mask);        // wait for SIGALRM or gc signal
	      setalarm();                     // switch thread on SIGALRM
	    }
	  lock();                             // prevent signal to be answered
	  preempt();                          // switch on SIGALRM signal
	}
    }
}

// ----------------------------------------------------------------------------
//
// block current thread where <mutex> is to be unlocked
// bring ready thread to execution
// if no ready thread found, realize lazy thread
// if no lazy thread found, switch to scheduler
// caller is in charge of locking the processor!!!
//
// ----------------------------------------------------------------------------
void
Processor::block(Mutex *mutex)
{
  CoreThread *ready = sys->get_ready();   // get ready thread
  if (!ready)                             // ready thread not found
    {                                  
      CoreThread *lazy = sys->get_lazy(); // get lazy thread
      if (lazy) lazy->realize();          // realize lazy thread
      ready = lazy;                       // new ready thread
    }
  save();                                 // going to switch thread
  if (ready)                              // ready thread found
    {
      ready->exec_block(mutex);           // switch to ready thread
    }
  else
    {
      sched->exec_block(mutex);           // switch to scheduler
    }
}

// ----------------------------------------------------------------------------
//
// poll for lazy thread and bring it to execution
// if no lazy thread found, switch to ready thread
// if no ready thread found, switch to scheduler thread
// in both cases, the current thread is deallocated
// otherwise, the current thread remains active!!!
// caller is in charge of locking the processor!!!
//
// ----------------------------------------------------------------------------
void
Processor::poll(void)
{
  while (1)
    {
      CoreThread *lazy = sys->get_lazy();     // get lazy thread
      if (!lazy) break;	                      // terminate curren thread
      lazy->next = 0;                         // push thread on lazy list
      thread->lazy = lazy;                    // critical region for gc!!!
      lazy->lazy = thread;
      unlock();                               // exit kernel mode
      thread->check();                        // check for stack overflow
      lazy->userfun(lazy);		      // execute lazy thread
      lazy->deliver();		              // and deliver result
      lock();                                 // lock processor again
      thread->lazy = 0;                       // pop thread from lazy list
      lazy->lazy = 0;
    }
  CoreThread *ready = sys->get_ready();       // get ready thread
  save();                                     // going to switch thread
  if (ready)	      	                      // ready thread was found
    {
      ready->exec_term();                     // switch to ready thread
    }
  else 
    {
      sched->exec_term();                     // switch to scheduler
    }
}

// ----------------------------------------------------------------------------
//
// unlink block of size <s> from current free block
//
// ----------------------------------------------------------------------------
inline Block *
Processor::unlink(int s)
{
  size -= s;                       // new block size
  Block *block = 		   
    (Block*)(((void*)free)+size);  // block from end of free 
  block->size = s;                 // size of block
  if (size == 0)                   // free block consumed
    {				   
      if (prev == free)            // free list empty
	{			   
	  free = 0;                // restore local data
	  prev = 0;		   
	}			   
      else                         // at least one block left
	{			   
	  free = free->next;       // take next block in cycle
	  prev->next = free;	   
	  size = free->size;       // adjust data structures
	}			   
    }				   
  return(block);                   // result block
}

// ----------------------------------------------------------------------------
//
// return memory block of at least <s> bytes unlinked from free list
// if not available, return 0
//
// ----------------------------------------------------------------------------
inline Block *
Processor::search(int s)
{
  if (size >= s) return(unlink(s));     // test current block
  if (!free) return(0);                 // no other block available
  free->size = size;                    // restore size information
  FreeBlock *start = free;              // start of search cycle
  while(1)			       
    {				       
      prev = free;                      // next block in cycle
      free = free->next;	       
      size = free->size;	       
      if (free == start) return(0);     // no more block available
      if (size >= s) return(unlink(s)); // test current block
    }
}

// ----------------------------------------------------------------------------
//
// allocate <s> bytes
// caller must unlock processor!!!
//
// ----------------------------------------------------------------------------
void*
Processor::alloc(int s)
{
  // alignment to header bound and to machine alignment
  s = (int)rt_align((void*)Block::align(sizeof(Block)+s));
  lock();                              // caller must unlock!
  Block *block = search(s);            // search for block in local list
  while (!block)                       // block not found
    {				       
      if (free) free->size = size;     // restore size information
      FreeBlock *next, *last;          // NOT free/prev (unlocking possible)!!!
      next = heap->alloc(s, &last);    // allocate new block list
      if (next)                        // allocation successful, still locked
	{			       
	  prev = last;                 // restore local data structures
	  free = next;		       
	  size = next->size;           
	  block = search(s);           // search again for block
	}			       
      else                             // processor was unlocked
	{			       
	  lock();                      // lock processor again
	  block = search(s);           // search for block in local list
	}			       
    }				
  return((void*)(block+1));            // address after block header
}

// ----------------------------------------------------------------------------
//
// prepare for garbage collection
//
// ----------------------------------------------------------------------------
void
Processor::gc(void)
{
  if (free) free->size = size; // restore size information
  size = 0;                    // empty free list
  free = 0;
  prev = 0;
}

// ----------------------------------------------------------------------------
// end of rt_proc.cc
// ----------------------------------------------------------------------------
// Local Variables: ***
// mode: C++ ***
// End: ***
