// ----------------------------------------------------------------------------
//
// rt_tbag.cc
// RT++ thread bags
//
// $Id: rt_tbag.cc,v 1.20 1996/03/22 13:52:38 ws Exp $
// (C) 1996 Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
//
// ----------------------------------------------------------------------------

#include "rt++.h"

Atom(CoreThread*);
Ref(Array<CoreThread*>);
Ref(List<CoreThread*>);

// ----------------------------------------------------------------------------
//
// wait until result is available
//
// ----------------------------------------------------------------------------
void 
ThreadBagCore::wait(int n)
{
  processor.lock();
  mutex.lock();
  if (n < results || (closed && n >= threads))
    {
      mutex.unlock();
      processor.unlock();
      return;
    }
  CoreThread *thread = processor.get_thread();
  if (n >= threads)
    {
      thread->next = wait_ahead;
      wait_ahead = thread;
    }
  else if (closed)
    {
      thread->next = wait_array[n];
      wait_array[n] = thread;
    }
  else
    {
      ThreadList *next = wait_list;
      for (int i = 0; i < n; i++)
	next = next->next;
      thread->next = next->list;
      next->list = thread;
    }
  processor.block(&mutex);
}

// ----------------------------------------------------------------------------
//
// return next result number
// processor and mutex remain locked!!!
//
// ----------------------------------------------------------------------------
int 
ThreadBagCore::prepare(void)
{
  processor.lock();
  mutex.lock();
  if (stopped)                // bag is stopped
    {
      processor.unlock();
      mutex.unlock();
      while (1) rt_yield();   // wait to die
    }
  return(results);
}

// ----------------------------------------------------------------------------
//
// awake thread list <list>
//
// ----------------------------------------------------------------------------
void
ThreadBagCore::awake(CoreThread *list)
{
  while (list)
    {
      CoreThread *thread = list;
      list = thread->next;
      sys->is_ready(thread);
    }
}

// ----------------------------------------------------------------------------
//
// awake waiting threads
// caller is in charge of locking processor and mutex!!!
// processor and mutex remain locked!!!
//
// ----------------------------------------------------------------------------
void
ThreadBagCore::awake(void)
{
  CoreThread *next = 0;
  if (closed)
    {
      next = wait_array[results];
      wait_array[results] = 0;
    }
  else
    {
      int n = results;
      ThreadList *list = wait_list;
      for (int i = 0; i < n; i++)
	list = list->next;
      next = list->list;
      list->list = 0;
    }
  awake(next);
  results++;
}

// ----------------------------------------------------------------------------
//
// awake waiting threads and unlink thread from bag
// caller is in charge of locking processor and mutex!!!
//
// ----------------------------------------------------------------------------
void
ThreadBagCore::deliver(CoreBagThread *thread)
{
  awake();
  CoreBagThread *_bnext = thread->bnext;
  CoreBagThread *_bprev = thread->bprev;
  if (_bnext)
    {
      _bnext->bprev = _bprev;
    }
  if (_bprev)
    {
      _bprev->bnext = _bnext;
    }
  else
    {
      tlist = _bnext;
    }
  mutex.unlock();
  processor.unlock();
}

// ----------------------------------------------------------------------------
//
// close bag with <k> more threads to be expected
// processor and mutex remain LOCKED!!!
//
// ----------------------------------------------------------------------------
void
ThreadBagCore::close(int k)
{
  int n = threads;
  wait_array = (CoreThread**)processor.alloc((n+k)*sizeof(CoreThread*));
  mutex.lock();
  for (int i = 0; i < n; i++)
    {
      wait_array[i] = wait_list->list;
      wait_list = wait_list->next;
    }
  for (int j = n; j < n+k; j++)
    wait_array[j] = 0;
  rest = k;
  threads += k;
  awake(wait_ahead);
  awake(wait_size);
  wait_ahead = 0;
  wait_size = 0;
  closed = 1;
}

// ----------------------------------------------------------------------------
//
// scan of list of linked threads
//
// ----------------------------------------------------------------------------
static void
scanthreads(CoreThread *thread)
{
  while (thread)
    {
      if (!Block::block(thread)->mark())
	thread->scanfun(thread);
      thread = thread->next;
    }
}

// ----------------------------------------------------------------------------
//
// scan of thread list
//
// ----------------------------------------------------------------------------
static void
scanlist(ThreadList *list)
{
  while (list)
    {
      Block::block(list)->mark();
      scanthreads(list->list);
      list = list->next;
    }
}

// ----------------------------------------------------------------------------
//
// scan of thread array
//
// ----------------------------------------------------------------------------
static void
scanarray(CoreThread **array, int n)
{
  Block::block(array)->mark();
  for (int i = 0; i < n; i++)
    scanthreads(array[i]);
}

// ----------------------------------------------------------------------------
//
// scan of local data structures
//
// ----------------------------------------------------------------------------
void
ThreadBagCore::scan(void)
{
  if (closed)
    scanarray(wait_array, threads);
  else
    scanlist(wait_list);
  scanthreads(wait_size);
  scanthreads(wait_ahead);
  CoreBagThread *thread = tlist;
  while (thread)
    {
      if (!Block::block(thread)->mark())
	thread->scanfun(thread);
      thread = thread->bnext;
    }
}

// ----------------------------------------------------------------------------
//
// initialization routine
//
// ----------------------------------------------------------------------------
void 
ThreadBagCore::init(void)
{
  mutex.init();
  closed = 0;
  rest = 0;
  stopped = 0;
  threads = 0;
  results = 0;
  tlist = 0;
  wait_list = 0;
  wait_tail = 0;
  wait_array = 0;
  wait_ahead = 0;
  wait_size = 0;
}

// ----------------------------------------------------------------------------
//
// add thread to bag
//
// ----------------------------------------------------------------------------
void
ThreadBagCore::add(CoreBagThread *thread)
{
  ThreadList *last = 0;
  if (closed)
    processor.lock();
  else
    {
      last = (ThreadList*)processor.alloc(sizeof(ThreadList));
      last->list = 0;
      last->next = 0;
    }
  mutex.lock();
  if (closed) 
    {
      if (rest == 0) rt_abort("bag is closed");
      rest--;
    }
  else
    {
      threads++;
      if (wait_tail)
	wait_tail->next = last;
      else
	wait_list = last;
      wait_tail = last;
    }
  CoreBagThread *next = tlist;
  if (next) next->bprev = thread;
  thread->bnext = next;
  thread->bprev = 0;
  tlist = thread;
  awake(wait_ahead);
  wait_ahead = 0;
  mutex.unlock();
  processor.unlock();
}

// ----------------------------------------------------------------------------
//
// return number of threads in bag, wait until bag is closed
//
// ----------------------------------------------------------------------------
int 
ThreadBagCore::size(void)
{
  if (closed) return(threads);
  processor.lock();
  mutex.lock();
  if (closed)
    {
      mutex.unlock();
      processor.unlock();
    }
  else
    {
      CoreThread *thread = processor.get_thread();
      thread->next = wait_size;
      wait_size = thread;
      processor.block(&mutex);
    }
  return(threads);
}

// ----------------------------------------------------------------------------
//
// stop threads associated to bag
//
// ----------------------------------------------------------------------------
void
ThreadBagCore::stop(void)
{
  processor.lock();
  mutex.lock();
  stopped = 1;
  mutex.unlock();
  processor.unlock();
  CoreBagThread *tl = tlist;
  while (tl)
    {
      CoreBagThread *t = tl;
      tl = tl->bnext;
      t->stop();
    }  
  tlist = 0;
}

// ----------------------------------------------------------------------------
// end of rt_tbag.cc
// ----------------------------------------------------------------------------
// Local Variables: ***
// mode: C++ ***
// End: ***
